import httpClient
import htmlparser
import xmltree
import asyncdispatch
import strutils
from uri import encodeURL, decodeURL
from os import paramCount, paramStr, getCurrentDir, dirExists, createDir, getHomeDir

type
  DorkFile = ref object
    title: string
    realURL: string
    cacheURL: string
  Options = ref object
    domain: string
    query: string
    extension: string
    path: string
    agent: string

var
  allDorkResults: seq[DorkFile]


proc showHelpCmd(keyword = "help", args = "", descr = "") =
  #[
    Make color for command syntax in help bannner
    Print them in help
    Syntax: <command> <keyword> <args> [<description>]
    command -> light green
    keyword -> red
    args (optional) -> yellow
    description (optional) -> blue
  ]#
  var cmdOutput = "  "
  cmdOutput &= "\e[31m" & keyword & "\e[0m " # Red color for keyword
  if args != "":
    cmdOutput &= "<\e[33m" & args & "\e[0m> "
  if descr != "":
    cmdOutput &= "[\e[94m" & descr & "\e[0m]"
  
  echo cmdOutput


proc banner() =
  #[
    Print core banner of program
  ]#
  echo "goofileN [0.2.1]: Filetype search with Google engine"
  echo "Developer: Nong Hoang \"DmKnght\" Tu"
  echo "Homepage: https://nest.parrot.sh/packages/tools/goofilen"


proc helpBanner() = 
  #[
    Print help commands with colored format
  ]#
  echo "\nUsage:"
  showHelpCmd(keyword = "-h | -help | --help | help", descr = "Show help banner")
  showHelpCmd(keyword = "-d", args = "Domain name", descr = "[Required] Target's name (ex: parrotsec.org)")
  showHelpCmd(keyword = "-f", args = "File extension", descr = "[Required] Extension without \".\" (ex: \"xlsx\", \"xlsx,pdf\")")
  showHelpCmd(keyword = "-q", args = "Query string", descr = "[Optional] Keywords to filter result (ex: members)")
  showHelpCmd(keyword = "-a", args = "User-Agent", descr = "[Optional] Use custom User-Agent")
  showHelpCmd(keyword = "-s", args = "Download folder", descr = "[Optional] Enable download files (ex: /tmp/)")


proc printResult() =
  #[
    Print result in table style
  ]#

  echo "-".repeat(70)
  for eachResult in allDorkResults:
    echo "| Title: " & "\e[31m" & eachResult.title & "\e[0m"
    echo "| URL: " & "\e[33m" & decodeURL(eachResult.realURL) & "\e[0m"
    echo "| Cache: " &  "\e[94m" & decodeURL(eachResult.cacheURL) & "\e[0m"
    echo "-".repeat(70)


proc downloadFiles(folder: string, agent: string) =
  #[
    Download file if users want to download all of them
    If URL can't be used, try use webcache
  ]#

  echo "[+] Downloading files from result"
  
  var
    client = newAsyncHttpClient()
    downloaded = 0

  client.headers = if agent == "": newHttpHeaders({"User-Agent": "Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)"}) else: newHttpHeaders({"User-Agent": agent})
  
  for eachResult in allDorkResults:
    let
      fileName = decodeUrl(eachResult.realURL).split("/")[^1]
      finalPath = if not contains(fileName, "&"): folder & fileName else: folder & fileName.split("&")[0]

    try:
      # Downloading -> yellow
      # Saved at -> cyan
      # Error: read
      let resp = waitFor(client.get(eachResult.realURL))
      if resp.status != "200 OK":
        if "&" in eachResult.realURL:
          let
            reTryURL = eachResult.realURL.split("&")[0]
            reTryResp = waitFor(client.get(reTryURL))
          if reTryResp.status != "200 OK":
            continue
          echo "[+] Downloading \e[33m" & reTryURL & "\e[0m"
          waitFor(client.downloadFile(reTryURL, finalPath))
          echo "[*] Saved at \e[96m" & finalPath & "\e[0m"
          downloaded += 1
      else:
        echo "[+] Downloading \e[33m" & eachResult.realURL & "\e[0m"
        waitFor(client.downloadFile(eachResult.realURL, finalPath))
        echo "[*] Saved at \e[96m" & finalPath & "\e[0m"
        downloaded += 1
    except:
      discard
      # stderr.write("[x] Error " & getCurrentExceptionMsg() & " \e[31m" & eachResult.realURL & "\e[0m\n")
      # if eachResult.cacheURL != "No web cache URL found":
      #   echo "[-] Download from cache is not supported. Please try it manually: " & eachResult.cacheURL
    
  echo "[*] Downloaded " & intToStr(downloaded) & " files"
  client.close()


proc getResult(resp: AsyncResponse, ext: string) =
  #[
    Check respose and create filter for the URL
    Expect result:
      1. Title
      2. URL
      3. Web cache URL 
  ]#

  if resp.status != "200 OK":
    stderr.write("[x] Get status code " & resp.status & "\n")
  else:
    let allResults = parseHTML(waitfor(resp.body))
    for eachResult in allResults.findAll("div"):
      # Filter all result thing
      let
        urlResults = eachResult.findAll("a")
        titleResults = eachResult.findAll("h3")
      var
        tmpCacheURL = ""
        dorkResult: DorkFile
      if urlResults == @[]:
        continue
      try:
        tmpCacheURL = urlResults[2].attr("href")
      except IndexDefect:
        tmpCacheURL = "No web cache URL found"
      
      try:
        dorkResult = DorkFile(
          title: titleResults[0].innerText,
          realURL: if urlResults[0].attr("href").startsWith("/url?q="): urlResults[0].attr("href")[ 7 .. ^1] else: urlResults[0].attr("href"),
          cacheURL: tmpCacheURL,
        )

        allDorkResults.add(dorkResult)
      except:
        # Google Dork has some false URL like /search?q=site:gov.vn ext:(xlsx | pdf) in resp
        # We ignore it
        discard


proc basicSearch(domain, extension, query, agent: string) =
  #[
    Do the google dork by basic HTTP GET queries
    Make the basic query like this:
      site:<domain name> ext:<file type> <Keyword>
    Num=500 how many results in first page
    Q= Our queries
  ]#
  var
    client = newAsyncHttpClient()
    payload = "site:" & domain & " ext:" & extension
  
  if query != "":
    payload &= " " & query
  
  payload = encodeUrl(payload)

  client.headers = if agent == "": newHttpHeaders({"User-Agent": "Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)"}) else: newHttpHeaders({"User-Agent": agent})
  let resp = client.get("https://www.google.com/search?num=500&q=" & payload)
  client.close()
  getResult(waitfor(resp), extension)


proc handleOptions(options: Options) =
  #[
    Handle all user's options and throw error
      1. Verify required options and run the basic search (no API support by now)
      2. If result > 0 print as table
      3. If result > 0 and user want to download: Download all file
  ]#
  if options.domain == "" or options.extension == "":
    stderr.write("[x] Domain name and file extension is required\n")
  else:
    if "," in options.extension:
      options.extension = "(" & options.extension.replace(",", " | ") & ")"
    basicSearch(options.domain, options.extension, options.query, options.agent)
    let szResults = len(allDorkResults)
    if szResults > 0:
      printResult()
      if options.path == ".":
        options.path = getCurrentDir()
      elif options.path.startsWith("~"):
        options.path = options.path.replace("~", getHomeDir())
      if options.path != "" and options.path[^1] != "/"[0]:
        options.path &= "/"

      if options.path != "":
        if not dirExists(options.path):
          try:
            createDir(options.path)
          except:
            stderr.write("[x] Error while creating download folder\n")
            return
          downloadFiles(options.path, options.agent)
        else:
          downloadFiles(options.path, options.agent)
    
    echo "Found " &  intToStr(szResults) & " results"


proc main =
  #[
    Handle arguments to pass to function
    1. Get all basic arguments for query
    2. Handle all aguments or print error / help
    3. Call job
    4 Print result
    5. Handle file download
  ]#

  #[
    All needed arguments:
      1. Target (domain)
      2. File type
      3. Query
      4. Optional: save folder
  ]#

  banner()

  if paramCount() == 0:
    helpBanner()
    return
  elif paramCount() == 1 and paramStr(1) == "help" or paramStr(1) == "-h" or paramStr(1) == "--help" or paramStr(1) == "-help":
    helpBanner()
    return
  else:
    var
      index = 1
      options: Options

    options = Options(
      domain: "",
      query: "",
      extension: "",
      path: "",
      agent: "",
    )

    while index < paramCount():
      if paramStr(index) == "-d":
        options.domain = paramStr(index + 1)
        index += 1
      elif paramStr(index) == "-f":
        options.extension = paramStr(index + 1)
        index += 1
      elif paramStr(index) == "-q":
        options.query = paramStr(index + 1)
        index += 1
      elif paramStr(index) == "-s":
        options.path = paramStr(index + 1)
        index += 1
      elif paramStr(index) == "-a":
        options.agent = paramStr(index + 1)
        index += 1
      else:
        stderr.write("[x] Unknow options " & paramStr(index) & "\n")
        helpBanner()
        return
      index += 1

    handleOptions(options)

    echo "[*] Completed"

main()
